/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#include "SpringDashPot.h"
using namespace std;

#include <cvode.h>
#include <nvector.h>
#include <cvdense.h>

#define TOLER 1e-8

void __stdcall SpringDashPotRHS(int n, real t, N_Vector y, N_Vector ydot, void *f_data);

double interpolate_lin(double x0, double y0, double x1, double y1, double x)
{
	double m = (y1 - y0) / (x1 - x0);
	return y1 + m*(x - x1);
}

void __stdcall SpringDashPotRHS(int n, real t, N_Vector y, N_Vector ydot, void *f_data)
{
	((SpringDashPot_Pager *) f_data)->RHS(n, t, y, ydot);
}

void SpringDashPot_Pager::RHS(int n, double t, void* _y, void* _ydot)
{
	N_Vector y = (N_Vector) _y;
	N_Vector ydot = (N_Vector) _ydot;

	double position = N_VIth(y, 0);
	double velocity = N_VIth(y, 1);

	// derivative of position is velocity
	N_VIth(ydot, 0) = velocity;

	// time-in-page to time-in-seg
	t += sampling * pageStart;
	
	// derivative of velocity is [(-2*z*w * velocity) - (w*w * (position - input))]
	double input; // from loader.data; most likely needs (linear) interpolation
	int ixp = (int) (floor(t/sampling) - pageStart); // index in page
	int actualPageStart = pageStart;
	double *data = loader.data;
	if ( (ixp < 0) || (ixp >= pageSize) ) {
		// CVODE ranges all over the data!
		ixp = min(dataCount-1, max(0, ixp+pageStart));
		xloader.loadPageContaining(ixp);
		actualPageStart = pageSize * (ixp / pageSize);
		ixp %= pageSize;
		data = xloader.data;
	}
	double ixp_t = (ixp + actualPageStart) * sampling;
	int ixp2; // second reference point for interpolation
	if ( ixp == 0 || (ixp != (pageSize-1) && t >= ixp_t) )
		ixp2 = ixp + 1;
	else
		ixp2 = ixp - 1;
	input = interpolate_lin(ixp_t, data[ixp],
							(ixp2 + pageStart) * sampling, data[ixp2],
							t);
	N_VIth(ydot, 1) = ((-2*damping*natfreq*velocity) - (natfreq*natfreq*(position-input)));
}

SpringDashPot_Pager::SpringDashPot_Pager( QUB_Tree segNode, QTR_NodeFptr loadDataCB, int pagesize, std::ostream& milerr, int samplesOffset, double samplingSec, double damping_ratio, double natural_freq )
	: QUB_Pager(((int *) segNode.data())[1] - ((int *) segNode.data())[0] + 1,
				pagesize, -pagesize, new double[pagesize]),
	  loader(segNode, loadDataCB, pagesize, milerr, samplesOffset),
	  xloader(segNode, loadDataCB, pagesize, milerr, samplesOffset),
	  sampling( samplingSec ),
	  damping( damping_ratio ),
	  natfreq( natural_freq ),
	  toler( TOLER ),
	  cvode_mem( NULL ),
	  y( NULL ),
	  t( - samplingSec )
{
	pageSize = loader.pageSize;
	pageStart = loader.pageStart;
	LoadCVode();
	LoadCVDense();
	loader.loadPageContaining(0);
	N_Vector yy;
	N_VMAKE(yy, new double[2], 2);
	y = yy;
	N_VIth(yy, 0) = loader.data[0];
	N_VIth(yy, 1) = 0.0;
	cvode_mem = CVodeMalloc(2, SpringDashPotRHS, 0.0, (N_Vector)y, BDF, NEWTON, SS, &toler, &toler, this, NULL, false, NULL, NULL, NULL); // ADAMS, FUNCTIONAL
}

SpringDashPot_Pager::~SpringDashPot_Pager()
{
	delete [] data;
	delete [] ((N_Vector)y)->data;
	N_VDISPOSE(((N_Vector)y));
	if ( cvode_mem )
		CVodeFree(cvode_mem);
	UnloadCVode();
}

void SpringDashPot_Pager::nextPage()
{
	pageStart += loader.pageSize;
	loader.loadPageContaining( pageStart );

	if ( cvode_mem == NULL ) {
		// just copy raw data
		memcpy(data, loader.data, loader.pageSize * sizeof(double));
		return;
	}

	N_VIth(((N_Vector)y), 0) = loader.data[0];
	double t0 = 0.0;
	double position = N_VIth(((N_Vector)y), 0);
	double velocity = N_VIth(((N_Vector)y), 1);
	int flag = CVReInit(cvode_mem, SpringDashPotRHS, t0, (N_Vector)y, BDF, NEWTON, SS, &toler, &toler, this, NULL, false, NULL, NULL, NULL); // ADAMS, FUNCTIONAL
	if ( flag ) {
		loader.err << "CVReInit in SpringDashPot_Pager" << endl;
		// just copy raw data
		memcpy(data, loader.data, loader.pageSize * sizeof(double));
		return;
	}

	CVDense(cvode_mem, NULL, NULL);

	double tRet;
	int N = loader.pageSize;
	for (int i=0; i<N; ++i) {
		if ( i ) { // not at t0
			flag = CVode(cvode_mem, i*sampling, (N_Vector)y, &tRet, NORMAL);
		}
		if ( flag ) {
			loader.err << "CVode: code " << flag << " in SpringDashPot_Pager" << endl;
			memcpy(data+i, loader.data+i, (loader.pageSize - i) * sizeof(double));
			break;
		} else {
			data[i] = N_VIth(((N_Vector)y), 0);
		}
	}

	if ( pageStart + pageSize < dataCount )
		CVode(cvode_mem, pageSize*sampling, (N_Vector)y, &tRet, NORMAL); // prep y for next page
}
